@@ -149,6 +149,7 @@ static void commandMsg(uint id, char *params) {
 }
 
 static void commandJoin(uint id, char *params) {
+	if (!params && id == Network) params = self.invited;
 	if (!params) params = idNames[id];
 	uint count = 1;
 	for (char *ch = params; *ch && *ch != ' '; ++ch) {
@@ -263,7 +264,7 @@ static void commandOp(uint id, char *params) {
 	if (params) {
 		channelListMode(id, '+', 'o', params);
 	} else {
-		ircFormat("PRIVMSG ChanServ :OP %s\r\n", idNames[id]);
+		ircFormat("CS OP %s\r\n", idNames[id]);
 	}
 }
 
@@ -275,7 +276,7 @@ static void commandVoice(uint id, char *params) {
 	if (params) {
 		channelListMode(id, '+', 'v', params);
 	} else {
-		ircFormat("PRIVMSG ChanServ :VOICE %s\r\n", idNames[id]);
+		ircFormat("CS VOICE %s\r\n", idNames[id]);
 	}
 }
 
@@ -355,12 +356,12 @@ static void commandWhowas(uint id, char *params) {
 
 static void commandNS(uint id, char *params) {
 	(void)id;
-	ircFormat("PRIVMSG NickServ :%s\r\n", (params ?: "HELP"));
+	ircFormat("NS %s\r\n", (params ?: "HELP"));
 }
 
 static void commandCS(uint id, char *params) {
 	(void)id;
-	ircFormat("PRIVMSG ChanServ :%s\r\n", (params ?: "HELP"));
+	ircFormat("CS %s\r\n", (params ?: "HELP"));
 }
 
 static void commandQuery(uint id, char *params) {
@@ -379,7 +380,17 @@ static void commandWindow(uint id, char *params) {
 		uiShowNum(strtoul(params, NULL, 10));
 	} else {
 		id = idFind(params);
-		if (id) uiShowID(id);
+		if (id) {
+			uiShowID(id);
+			return;
+		}
+		for (const char *match; (match = completeSubstr(None, params));) {
+			id = idFind(match);
+			if (!id) continue;
+			completeAccept();
+			uiShowID(id);
+			break;
+		}
 	}
 }
 
@@ -472,6 +483,7 @@ static void commandExec(uint id, char *params) {
 	if (pid < 0) err(EX_OSERR, "fork");
 	if (pid) return;
 
+	setsid();
 	close(STDIN_FILENO);
 	dup2(execPipe[1], STDOUT_FILENO);
 	dup2(utilPipe[1], STDERR_FILENO);
@@ -529,7 +541,7 @@ static const struct Handler {
 	{ "/deop", commandDeop, 0 },
 	{ "/devoice", commandDevoice, 0 },
 	{ "/except", commandExcept, 0 },
-	{ "/exec", commandExec, Multiline | Restrict },
+	{ "/exec", commandExec, Multiline | Restrict | Kiosk },
 	{ "/help", commandHelp, 0 }, // Restrict special case.
 	{ "/highlight", commandHighlight, 0 },
 	{ "/ignore", commandIgnore, 0 },
