@@ -32,6 +32,7 @@
 #include <curses.h>
 #include <err.h>
 #include <errno.h>
+#include <fcntl.h>
 #include <limits.h>
 #include <signal.h>
 #include <stdarg.h>
@@ -40,6 +41,7 @@
 #include <stdlib.h>
 #include <string.h>
 #include <sysexits.h>
+#include <sys/file.h>
 #include <term.h>
 #include <termios.h>
 #include <time.h>
@@ -47,6 +49,10 @@
 #include <wchar.h>
 #include <wctype.h>
 
+#ifdef __FreeBSD__
+#include <capsicum_helpers.h>
+#endif
+
 #include "chat.h"
 
 // Annoying stuff from <term.h>:
@@ -122,6 +128,8 @@ static struct Window *windowRemove(uint num) {
 	return window;
 }
 
+enum Heat uiThreshold = Cold;
+
 static uint windowFor(uint id) {
 	for (uint num = 0; num < windows.len; ++num) {
 		if (windows.ptrs[num]->id == id) return num;
@@ -131,7 +139,11 @@ static uint windowFor(uint id) {
 	window->id = id;
 	window->mark = true;
 	window->time = uiTime.enable;
-	window->thresh = Cold;
+	if (id == Network || id == Debug) {
+		window->thresh = Cold;
+	} else {
+		window->thresh = uiThreshold;
+	}
 	window->buffer = bufferAlloc();
 	completeAdd(None, idNames[id], idColors[id]);
 	return windowPush(window);
@@ -374,14 +386,16 @@ static short stylePair(struct Style style) {
 	return colorPair(Colors[style.fg], Colors[style.bg]);
 }
 
-static void styleAdd(WINDOW *win, const char *str) {
+static int styleAdd(WINDOW *win, const char *str) {
 	struct Style style = StyleDefault;
 	while (*str) {
 		size_t len = styleParse(&style, &str);
 		wattr_set(win, styleAttr(style), stylePair(style), NULL);
-		waddnstr(win, str, len);
+		if (waddnstr(win, str, len) == ERR)
+			return -1;
 		str += len;
 	}
+	return 0;
 }
 
 static void statusUpdate(void) {
@@ -401,37 +415,42 @@ static void statusUpdate(void) {
 			others.unread += window->unreadWarm;
 			if (window->heat > others.heat) others.heat = window->heat;
 		}
-		char buf[256] = "";
-		struct Cat cat = { buf, sizeof(buf), 0 };
-		catf(
-			&cat, "\3%d%s %u%s%s %s ",
+		char buf[256], *end = &buf[sizeof(buf)];
+		char *ptr = seprintf(
+			buf, end, "\3%d%s %u%s%s %s ",
 			idColors[window->id], (num == windows.show ? "\26" : ""),
 			num, window->thresh[(const char *[]) { "-", "", "+", "++" }],
 			&"="[!window->mute], idNames[window->id]
 		);
 		if (window->mark && window->unreadWarm) {
-			catf(
-				&cat, "\3%d+%d\3%d%s",
+			ptr = seprintf(
+				ptr, end, "\3%d+%d\3%d%s",
 				(window->heat > Warm ? White : idColors[window->id]),
 				window->unreadWarm, idColors[window->id],
 				(window->scroll ? "" : " ")
 			);
 		}
 		if (window->scroll) {
-			catf(&cat, "~%d ", window->scroll);
+			ptr = seprintf(ptr, end, "~%d ", window->scroll);
 		}
-		styleAdd(status, buf);
+		if (styleAdd(status, buf) < 0) break;
 	}
 	wclrtoeol(status);
 
-	struct Cat cat = { title, sizeof(title), 0 };
 	const struct Window *window = windows.ptrs[windows.show];
-	catf(&cat, "%s %s", network.name, idNames[window->id]);
+	char *end = &title[sizeof(title)];
+	char *ptr = seprintf(
+		title, end, "%s %s", network.name, idNames[window->id]
+	);
 	if (window->mark && window->unreadWarm) {
-		catf(&cat, " +%d%s", window->unreadWarm, &"!"[window->heat < Hot]);
+		ptr = seprintf(
+			ptr, end, " +%d%s", window->unreadWarm, &"!"[window->heat < Hot]
+		);
 	}
 	if (others.unread) {
-		catf(&cat, " (+%d%s)", others.unread, &"!"[others.heat < Hot]);
+		ptr = seprintf(
+			ptr, end, " (+%d%s)", others.unread, &"!"[others.heat < Hot]
+		);
 	}
 }
 
@@ -548,8 +567,8 @@ static void notify(uint id, const char *str) {
 	if (self.restricted) return;
 	if (!uiNotifyUtil.argc) return;
 
-	char buf[1024] = "";
-	styleStrip(&(struct Cat) { buf, sizeof(buf), 0 }, str);
+	char buf[1024];
+	styleStrip(buf, sizeof(buf), str);
 
 	struct Util util = uiNotifyUtil;
 	utilPush(&util, idNames[id]);
@@ -559,6 +578,7 @@ static void notify(uint id, const char *str) {
 	if (pid < 0) err(EX_OSERR, "fork");
 	if (pid) return;
 
+	setsid();
 	close(STDIN_FILENO);
 	dup2(utilPipe[1], STDOUT_FILENO);
 	dup2(utilPipe[1], STDERR_FILENO);
@@ -1111,7 +1131,7 @@ static size_t signatureVersion(time_t signature) {
 	for (size_t i = 0; i < ARRAY_LEN(Signatures); ++i) {
 		if (signature == Signatures[i]) return i;
 	}
-	err(EX_DATAERR, "unknown file signature %jX", (uintmax_t)signature);
+	errx(EX_DATAERR, "unknown file signature %jX", (uintmax_t)signature);
 }
 
 static int writeTime(FILE *file, time_t time) {
@@ -1121,41 +1141,41 @@ static int writeString(FILE *file, const char *str) {
 	return (fwrite(str, strlen(str) + 1, 1, file) ? 0 : -1);
 }
 
-int uiSave(const char *name) {
-	FILE *file = dataOpen(name, "w");
-	if (!file) return -1;
+static FILE *saveFile;
 
+int uiSave(void) {
 	int error = 0
-		|| writeTime(file, Signatures[7])
-		|| writeTime(file, self.pos);
+		|| ftruncate(fileno(saveFile), 0)
+		|| writeTime(saveFile, Signatures[7])
+		|| writeTime(saveFile, self.pos);
 	if (error) return error;
 	for (uint num = 0; num < windows.len; ++num) {
 		const struct Window *window = windows.ptrs[num];
 		error = 0
-			|| writeString(file, idNames[window->id])
-			|| writeTime(file, window->mute)
-			|| writeTime(file, window->time)
-			|| writeTime(file, window->thresh)
-			|| writeTime(file, window->heat)
-			|| writeTime(file, window->unreadSoft)
-			|| writeTime(file, window->unreadWarm);
+			|| writeString(saveFile, idNames[window->id])
+			|| writeTime(saveFile, window->mute)
+			|| writeTime(saveFile, window->time)
+			|| writeTime(saveFile, window->thresh)
+			|| writeTime(saveFile, window->heat)
+			|| writeTime(saveFile, window->unreadSoft)
+			|| writeTime(saveFile, window->unreadWarm);
 		if (error) return error;
 		for (size_t i = 0; i < BufferCap; ++i) {
 			const struct Line *line = bufferSoft(window->buffer, i);
 			if (!line) continue;
 			error = 0
-				|| writeTime(file, line->time)
-				|| writeTime(file, line->heat)
-				|| writeString(file, line->str);
+				|| writeTime(saveFile, line->time)
+				|| writeTime(saveFile, line->heat)
+				|| writeString(saveFile, line->str);
 			if (error) return error;
 		}
-		error = writeTime(file, 0);
+		error = writeTime(saveFile, 0);
 		if (error) return error;
 	}
 	return 0
-		|| writeString(file, "")
-		|| urlSave(file)
-		|| fclose(file);
+		|| writeString(saveFile, "")
+		|| urlSave(saveFile)
+		|| fclose(saveFile);
 }
 
 static time_t readTime(FILE *file) {
@@ -1172,51 +1192,57 @@ static ssize_t readString(FILE *file, char **buf, size_t *cap) {
 }
 
 void uiLoad(const char *name) {
-	FILE *file = dataOpen(name, "r");
-	if (!file) {
-		if (errno != ENOENT) exit(EX_NOINPUT);
-		file = dataOpen(name, "w");
-		if (!file) exit(EX_CANTCREAT);
-		fclose(file);
-		return;
+	int error;
+	saveFile = dataOpen(name, "a+e");
+	if (!saveFile) exit(EX_CANTCREAT);
+	rewind(saveFile);
+
+#ifdef __FreeBSD__
+	cap_rights_t rights;
+	cap_rights_init(&rights, CAP_READ, CAP_WRITE, CAP_FLOCK, CAP_FTRUNCATE);
+	error = caph_rights_limit(fileno(saveFile), &rights);
+	if (error) err(EX_OSERR, "cap_rights_limit");
+#endif
+
+	error = flock(fileno(saveFile), LOCK_EX | LOCK_NB);
+	if (error && errno == EWOULDBLOCK) {
+		errx(EX_CANTCREAT, "%s: save file in use", name);
 	}
 
 	time_t signature;
-	fread(&signature, sizeof(signature), 1, file);
-	if (ferror(file)) err(EX_IOERR, "fread");
-	if (feof(file)) {
-		fclose(file);
+	fread(&signature, sizeof(signature), 1, saveFile);
+	if (ferror(saveFile)) err(EX_IOERR, "fread");
+	if (feof(saveFile)) {
 		return;
 	}
 	size_t version = signatureVersion(signature);
 
 	if (version > 1) {
-		self.pos = readTime(file);
+		self.pos = readTime(saveFile);
 	}
 
 	char *buf = NULL;
 	size_t cap = 0;
-	while (0 < readString(file, &buf, &cap) && buf[0]) {
+	while (0 < readString(saveFile, &buf, &cap) && buf[0]) {
 		struct Window *window = windows.ptrs[windowFor(idFor(buf))];
-		if (version > 3) window->mute = readTime(file);
-		if (version > 6) window->time = readTime(file);
-		if (version > 5) window->thresh = readTime(file);
+		if (version > 3) window->mute = readTime(saveFile);
+		if (version > 6) window->time = readTime(saveFile);
+		if (version > 5) window->thresh = readTime(saveFile);
 		if (version > 0) {
-			window->heat = readTime(file);
-			window->unreadSoft = readTime(file);
-			window->unreadWarm = readTime(file);
+			window->heat = readTime(saveFile);
+			window->unreadSoft = readTime(saveFile);
+			window->unreadWarm = readTime(saveFile);
 		}
 		for (;;) {
-			time_t time = readTime(file);
+			time_t time = readTime(saveFile);
 			if (!time) break;
-			enum Heat heat = (version > 2 ? readTime(file) : Cold);
-			readString(file, &buf, &cap);
+			enum Heat heat = (version > 2 ? readTime(saveFile) : Cold);
+			readString(saveFile, &buf, &cap);
 			bufferPush(window->buffer, COLS, window->thresh, heat, time, buf);
 		}
 		windowReflow(window);
 	}
-	urlLoad(file, version);
+	urlLoad(saveFile, version);
 
 	free(buf);
-	fclose(file);
 }
